# -*- coding: utf-8 -*-
"""
/***************************************************************************
 pointShifter
                                 A QGIS plugin
 pointShifter
                              -------------------
        begin                : 2016-01-21
        copyright            : (C) 2016 by Eugene Nazarenko
        email                : eugene.nazarenko@here.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
# Import the PyQt and QGIS libraries
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from qgis.core import *
from qgis.core import QgsRelation, QgsProject
from qgis.utils import plugins
from qgis.gui import QgsEncodingFileDialog
# Initialize Qt resources from file resources.py
import resources_rc
# Import the code for the dialog
from pointShifterdialog import pointShifterDialog
import os, os.path
import uuid
import math

class trace:
    """
    class for tracing debug infos
    """

    def __init__(self):
        self.trace = False

    def ce(self,string):
        if self.trace:
            print string

class pointShifter():

    def __init__(self, iface):
        # Save reference to the QGIS interface
        self.iface = iface
        # initialize plugin directory
        self.plugin_dir = os.path.dirname(__file__)
        # initialize locale
        locale = QSettings().value("locale/userLocale")[0:2]
        localePath = os.path.join(self.plugin_dir, 'i18n', 'pointShifter_{}.qm'.format(locale))

        if os.path.exists(localePath):
            self.translator = QTranslator()
            self.translator.load(localePath)

            if qVersion() > '4.3.3':
                QCoreApplication.installTranslator(self.translator)

        # Create the dialog (after translation) and keep reference
        self.dlg = pointShifterDialog()
        self.tra = trace()
        self.outFilePath = ''
        self.encoding = ''

    def initGui(self):
        # Create action that will start plugin configuration
        self.action = QAction(
            QIcon(":/plugins/pointShifter/icon.png"),
            u"pointShifter", self.iface.mainWindow())
        # connect the action to the run method
        self.action.triggered.connect(self.run)

        # Add toolbar button and menu item
        self.iface.addToolBarIcon(self.action)
        self.iface.addPluginToMenu(u"&pointShifter", self.action)
#        self.dlg.pointLayerCombo.activated.connect(self.browsePointInput)
#        self.dlg.lineLayerCombo.activated.connect(self.browseLineInput)

        self.dlg.browsePushButton.clicked.connect(self.browseOutputFC)
        self.dlg.buttonBox.accepted.connect(self.applyShift)

    def unload(self):
        # Remove the plugin menu item and icon
        self.iface.removePluginMenu(u"&pointShifter", self.action)
        self.iface.removeToolBarIcon(self.action)

    def populateComboBox(self,combo,list,predef = '',sort = True):
        #procedure to fill specified combobox with provided list
        combo.clear()
        model=QStandardItemModel(combo)
        for elem in list:
            try:
                item = QStandardItem(unicode(elem))
            except TypeError:
                item = QStandardItem(str(elem))
            model.appendRow(item)
        if sort:
            model.sort(0)

        combo.setModel(model)
        if predef != "":
            for row in range (0,combo.count()):
                if combo.itemText(row) == predef:
                    pos = row
            try:
                combo.setCurrentIndex(pos)
            except:
                combo.insertItem(0,predef)
                combo.setCurrentIndex(0)

    def browseOutputFC(self):
        filtering="Shapefiles (*.shp *.SHP)"
        settings = QSettings()
        dirName = settings.value("/UI/lastShapefileDir")
        encode = settings.value("/UI/encoding")
        fileDialog = QgsEncodingFileDialog(None, QCoreApplication.translate("fTools", "Save output shapefile"), dirName, filtering, encode)
        fileDialog.setDefaultSuffix("shp")
        fileDialog.setFileMode(QFileDialog.AnyFile)
        fileDialog.setAcceptMode(QFileDialog.AcceptSave)
        fileDialog.setConfirmOverwrite(True)
        if not fileDialog.exec_() == QDialog.Accepted:
            return None, None
        files = fileDialog.selectedFiles()
        settings.setValue("/UI/lastShapefileDir", QFileInfo(unicode(files[0])).absolutePath())

        #return (unicode(files[0]), unicode(fileDialog.encoding()))
        self.outFilePath = unicode(files[0])
        self.encoding = unicode(fileDialog.encoding())
        self.dlg.outputLineEdit.setText(self.outFilePath)




    # run method that performs all the real work
    def run(self):

##        if not 'refFunctions' in plugins:
##            QMessageBox.critical(None, "Plugin Dependency:", "Plugin 'refFunctions' is needed.\nInstall it from Qgis repository before performing spatial joins")
##            return

        self.pointLayerSet = {}
        for layer in self.iface.legendInterface().layers():
            if layer.type() == QgsMapLayer.VectorLayer and layer.geometryType() == QGis.Point:
                self.pointLayerSet[layer.name()]=layer

        self.lineLayerSet = {}
        for layer in self.iface.legendInterface().layers():
            if layer.type() == QgsMapLayer.VectorLayer and layer.geometryType() == QGis.Line:
                self.lineLayerSet[layer.name()]=layer


        self.dlg.progressBar.reset()

        cPointLayerName = "Please add point layer to map"
        if self.iface.legendInterface().currentLayer():
            if self.iface.legendInterface().currentLayer().geometryType() == QGis.Point:
                cPointLayerName = self.iface.legendInterface().currentLayer().name()
            else:
                cPointLayerName = self.pointLayerSet.keys()[0]


        cLineLayerName = "Please add line layer to map"
        if self.iface.legendInterface().currentLayer():
            if self.iface.legendInterface().currentLayer().geometryType() == QGis.Line:
                cLineLayerName = self.iface.legendInterface().currentLayer().name()
            else:
                cLineLayerName = self.lineLayerSet.keys()[0]


        self.populateComboBox(self.dlg.pointLayerCombo,self.pointLayerSet.keys(),cPointLayerName)
        self.populateComboBox(self.dlg.lineLayerCombo,self.lineLayerSet.keys(),cLineLayerName)
        # show the dialog
        self.dlg.show()


    def get_moved_point(self, x1, y1, x2, y2, side, l=0.0000090):
        from math import sqrt
        #l = 0.0000090 #0.5 meters
        len_current_vector = sqrt(pow(x2-x1, 2)+pow(y2-y1, 2))
        coeff = l/len_current_vector
        x3 = y3 = 0.0000000
        if side == 'R':
            x3 =  x2 - (y1-y2)
            y3 =  y2 + (x1-x2)
        elif side == 'L':
            x3 =  x2 + (y1-y2)
            y3 =  y2 - (x1-x2)
        outX = x2 - coeff*(x2 - x3)
        outY = y2 - coeff*(y2 - y3)
        return outX, outY

    def get_point_side_from_link (self, startX, startY, pointX, pointY, endX, endY):
        return (pointX-startX)*(pointY-endY)-(pointY-startY)*(pointX-endX)

    def get_point_on_link (self, x1, y1, x2, y2, px, py):
        from math import sqrt
        class Vector:
            def __init__(self, xcoord = 0.0, ycoord =0.0):
                self.x = xcoord
                self.y = ycoord

        def VDot(v1,v2): #Vector
            return (v1.x*v2.x+v1.y*v2.y)


        def VMul(v1, A): #Vector
            return Vector(v1.x*A, v1.y*A)



        def VSub(v1,v2): #Vector;
            return Vector(v1.x-v2.x, v1.y-v2.y)

        def VLength(V):
            return sqrt(V.x ** 2 + V.y ** 2)

        def VNorm(V): #Vector;
            #vl = 0.0
            vl = VLength(V)
            return Vector(V.x/vl, V.y/vl)


        def VProject(A,B): #Vector;
            An = VNorm(A)
            return VMul(An,VDot(An,B))

        def Perpendicular(A,B,C): #Vector;
            CA = Vector()
            CA=VSub(C,A)
            return VProject(VSub(B,A),CA)

        result = Perpendicular (Vector(x1, y1), Vector(x2, y2), Vector(px, py))
        return x1 + result.x, y1 + result.y


# main algorithm
    def applyShift(self):
        self.dlg.show()
        if self.dlg.pointLayerCombo.currentText()[:10] != 'Please add' and self.dlg.lineLayerCombo.currentText()[:10] != 'Please add':
            #prepare input layers
            pointLayer = self.pointLayerSet[self.dlg.pointLayerCombo.currentText()]
            lineLayer = self.lineLayerSet[self.dlg.lineLayerCombo.currentText()]
            self.dlg.progressBar.setMinimum(0)
            self.dlg.progressBar.setMaximum(pointLayer.featureCount())

            #prepare output layer

##            outputPointsShapeLyr.startEditing()
##            outputPointsShape = outputPointsShapeLyr.dataProvider()

##            for pntField in pointLayer.fields():
##                self.tra.ce(str(type(pntField)))
##                outputPointsShape.addAttribute(pntField)
##            outputPointsShape.addAttribute(QgsField('lat_f', QVariant.Double, 'double'))
##            outputPointsShape.addAttribute(QgsField('lon_f', QVariant.Double, 'double'))
##            outputPointsShape.addAttribute(QgsField('LINKPVID', QVariant.Int, 'integer'))
##            #outputPointsShape.addAttribute(QgsField('side', QVariant.String, 'text'))
##            outputPointsShape.addAttribute(QgsField('pfrn_d', QVariant.Double, 'double'))
##            outputPointsShape.addAttribute(QgsField('pfrn', QVariant.Int, 'integer'))
            sRs = pointLayer.crs()
            outFields = QgsFields()
            for pntField in pointLayer.fields():
                #self.tra.ce(str(type(pntField)))
                outFields.append(pntField)

            outFields.append(QgsField(u'lat_f', QVariant.Double))
            outFields.append(QgsField(u'lon_f', QVariant.Double))
            outFields.append(QgsField(u'LINKPVID', QVariant.Int))
            #outFields.append(QgsField('side', QVariant.String, 'text'))
            outFields.append(QgsField(u'pfrn_d', QVariant.Double))
            outFields.append(QgsField(u'pfrn', QVariant.Int))

            outputPointsShape = QgsVectorFileWriter(self.outFilePath, self.encoding, outFields, QGis.WKBPoint, sRs)

            #self.tra.ce(str(type(outputPointsShape)))


            #targetLayerFields = [field.name() for field in targetLayer.pendingFields()]
            try:
                self.tra.ce('setup layers for processing.')
                pointFeatures = pointLayer.getFeatures()
                lineFeatures = lineLayer.getFeatures()
                # go through all points
                sindex = QgsSpatialIndex()
                inFeat = QgsFeature()
                self.tra.ce('create Spatial Index object and fill it')
                while lineFeatures.nextFeature(inFeat):
                    sindex.insertFeature(inFeat)
                pntIndex = 0
                self.dlg.progressBar.reset()

                self.tra.ce('start processing')
                for pnt in pointFeatures:
                    #get point geometry object
                    pntGeom = pnt.geometry()
                    pnt_xy = pntGeom.asPoint() #QgsPoint type
                    #find nearest 5 links to it (5 cause this is not a fact that one link will be closest)
                    lineID = sindex.nearestNeighbor(pnt_xy,5)
                    self.tra.ce('nearest line id %s'%lineID)
                    linesRequest = QgsFeatureRequest()
                    linesRequest.setFilterFids(lineID)

                    lineFeatIter = lineLayer.getFeatures(linesRequest)

                    lineFeat = QgsFeature()
                    # find closests link to point
                    min_dist = 1.0
                    while lineFeatIter.nextFeature(lineFeat):
                        if lineFeat.geometry().distance(pntGeom) <= min_dist:
                            min_dist = lineFeat.geometry().distance(pntGeom)
                            lineID = lineFeat.id()

                    lineFeatIter = lineLayer.getFeatures(QgsFeatureRequest(lineID))
                    lineFeatIter.nextFeature(lineFeat)


                    lineGeom = lineFeat.geometry()
                    pntOnLine = QgsPoint(0,0) # future point on line
                    afterVertex = 0
                    pntTrueSide = 0
                    closeSegResult = lineGeom.closestSegmentWithContext(pnt_xy) #, pntOnLine, afterVertex, pntTrueSide values go to the closeSegResult
                    self.tra.ce('closestSegment.. method result\n%s'%str(closeSegResult))

                    #pntOnLine = closeSegResult[1]
                    afterVertex = closeSegResult[2]
                    #pntTrueSide = closeSegResult[3]

                    #    def get_point_on_link (x1, y1, x2, y2, px, py):
                    x,y = self.get_point_on_link(lineGeom.vertexAt(afterVertex-1).x(), lineGeom.vertexAt(afterVertex-1).y(), lineGeom.vertexAt(afterVertex).x(), lineGeom.vertexAt(afterVertex).y(), pnt_xy.x(), pnt_xy.y())
                    pntOnLine = QgsPoint(x, y)



                    vertexIndex = 0
                    lineLength2Point = 0.0
                    while lineGeom.vertexAt(vertexIndex) <> QgsPoint(0,0) and vertexIndex>0:
                        vertexIndex+=1
                        if vertexIndex < afterVertex:
                            lineLength2Point += math.sqrt(math.pow((lineGeom.vertexAt(vertexIndex).y()-lineGeom.vertexAt(vertexIndex-1).y()), 2)+pow((lineGeom.vertexAt(vertexIndex).x()-lineGeom.vertexAt(vertexIndex-1).x()), 2))
                    vertexIndex = vertexIndex-1

                    startLinePnt = lineGeom.vertexAt(0)
                    endLinePoint = lineGeom.vertexAt(vertexIndex)
                    #remark: code for define ref point of link
                    self.tra.ce('origin point %s, %s'%(pntOnLine.x(),pntOnLine.y()))

                    start_point_is_refnode = (startLinePnt.y() < endLinePoint.y() and startLinePnt.y() <> endLinePoint.y()) or (startLinePnt.x() < endLinePoint.x() and startLinePnt.x() <> endLinePoint.x())

                    self.tra.ce('Start point is ref node: %s'%start_point_is_refnode)

                    stSide = pnt.attribute('SIDE')
                    self.tra.ce('SIDE should be %s'%stSide)
                    #calc shifted point
                    self.tra.ce('point is shifting to %s map units'% float(self.dlg.lineEdit.text()))
                    #x,y = self.get_moved_point(lineGeom.vertexAt(afterVertex-1).x(), lineGeom.vertexAt(afterVertex-1).y(), pntOnLine.x(), pntOnLine.y(), stSide, float(self.dlg.lineEdit.text()))
                    x,y = self.get_moved_point(lineGeom.vertexAt(afterVertex-1).x(), lineGeom.vertexAt(afterVertex-1).y(), pntOnLine.x(), pntOnLine.y(), stSide, float(self.dlg.lineEdit.text()))

                    self.tra.ce('origin point %s, %s'%(x,y))
                    #OUTPUT
                    shiftedPoint = QgsPoint(x,y)

                    #calc percent from refnode
                    lastSegPartialLen = math.sqrt(math.pow((lineGeom.vertexAt(afterVertex-1).y()-pntOnLine.y()), 2)+pow((lineGeom.vertexAt(afterVertex-1).x()-pntOnLine.x()), 2))
                    percentFromStart = ((lineLength2Point+lastSegPartialLen)*100)/lineGeom.length()

                    self.tra.ce('Percent from ref node: %s'%percentFromStart)
                    if start_point_is_refnode:
                        #OUTPUT
                        percentFromStart = 100 - percentFromStart

                    #OUTPUT
                    linkPVID = lineFeat.attribute('LINK_ID')
                    self.tra.ce('LINK_ID: %s'%linkPVID)

                    self.tra.ce(str(pointLayer.fields().allAttributesList()))
                    #store procedure
                    currentFeature = QgsFeature(outFields)
                    currentFeature.initAttributes(outFields.count())

                    self.tra.ce(outFields.count())
                    self.tra.ce(currentFeature.fields().count())

                    attributes = pnt.attributes()
                    attributes += [shiftedPoint.y(), shiftedPoint.x(), linkPVID, percentFromStart, int(percentFromStart)]

                    self.tra.ce(attributes)

                    currentFeature.setGeometry(QgsGeometry.fromPoint(shiftedPoint))
                    #currentFeature.setGeometry(QgsGeometry.fromPoint(pntOnLine))
                    currentFeature.setAttributes(attributes)

                    self.tra.ce(currentFeature.attributes())

                    outputPointsShape.addFeature(currentFeature)

                    del currentFeature

                    #update progressBar
                    self.dlg.progressBar.setValue(pntIndex)
                    pntIndex+=1
            finally:
                del outputPointsShape



            outputPointsShapeLyr = QgsVectorLayer(self.outFilePath, u'ShiftedPoints', 'ogr')  #add sRs here (it is from point layer)
            QgsMapLayerRegistry.instance().addMapLayers([outputPointsShapeLyr])

            self.dlg.hide()


